@@ -143,11 +143,6 @@ main (int argc, char **argv)
     i = 1;
 #endif
 
-    if (!FcInit ())
-    {
-	fprintf (stderr, "Can't init font config library\n");
-	return 1;
-    }
     if (argv[i])
     {
 	pat = FcNameParse ((FcChar8 *) argv[i]);
@@ -230,7 +225,7 @@ main (int argc, char **argv)
 		if (s)
 		{
 		    printf ("%s", s);
-		    free (s);
+		    FcStrFree (s);
 		}
 	    }
 
